#include <thread>
#include <mutex>
#include <vector>
#include <chrono>
#include <cassert>
#include <iostream>

#include "test.h"

namespace test {

    void Reader::operator()() {
        ReadLock readLock( _lock );
        bool end = false;

        do {
            int A, B, C;
            {
                std::unique_lock< ReadLock > l( readLock );
                A = _data.A;
                B = _data.B;
                C = _data.C;
            }
            end = A == B && B == C && C == 0;
            assert( A - B == C );
        } while ( !end );
    }
    
    void Writer::operator()( int id ) {
        WriteLock writeLock( _lock );
        int init = 100 * id;
        for ( int i = init; i < init + 10; ++i ) {
            {
                std::unique_lock< WriteLock > l( writeLock );
                _data.A = i * 42;
                _data.B = _data.A * 13 + 42;
                _data.C = _data.A - _data.B;
                std::cout << "zapis " << id << std::endl;
            }
            std::chrono::milliseconds s( 700 );
            std::this_thread::sleep_for( s );
        }
    }
    
    void run() {
        Data data( 1, 1, 0 );
        RWlock rwlock;

        Reader reader( rwlock, data );
        Writer writer( rwlock, data );
    
        std::vector< std::thread > threads{ 10 };

        for ( auto &t : threads )
            t = std::thread( reader );
        auto w = std::thread( writer, 1 );

        std::chrono::seconds s( 1 );
        std::this_thread::sleep_for( s );
        writer( 2 );
        w.join();
        
        {
            WriteLock writeLock( rwlock );
            std::unique_lock< WriteLock > l( writeLock );
            data.A = data.B = data.C = 0;
        }
        
        for ( auto &t : threads )
            t.join();
    }
}


